import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Vector;

public class ArrayBasedQueue<E> implements QueueInterface<E> {

	private final int MAXIMUM_ARRAY_SIZE = 10_000;
	private E[] array = (E[]) new Object[MAXIMUM_ARRAY_SIZE];
	private int size = 0;
	private int front = -1;
	private int back  = -1;
	
	
	@Override
	public Iterator<E> iterator() {
		Vector<E> elements = new Vector<E>(this.size);
		
		for ( int i = 0 ; i < this.size ; i++ ) {
			int index = findIndex(i);
			elements.add(this.array[index]);
		}
		
		ElementIterator<E> elementIterator = new ElementIterator<E>(elements);
		return elementIterator;
	}

	@Override
	public boolean isEmpty() {
		return (this.size == 0 ? true : false);
	}

	@Override
	public int size() {
		return this.size;
	}

	@Override
	public void enqueue(E e) throws IllegalStateException, NullPointerException {
		if (e == null) {
			throw new NullPointerException("Unable to add a null pointer to the ArrayBasedQueue");
		} else if (this.size == this.array.length) {
			throw new IllegalStateException("The array is full, unable to add more elements");
		} 
		
		if (this.isEmpty()) {
			array[0] = e;
			this.front = 0;
			this.back = 0;
		} else {
			this.back = this.incrementIndex(this.back);
			this.array[this.back] = e;
		}
		
		this.size++;
	}

	@Override
	public E peek() {
		if (this.isEmpty()) {
			return null;
		} else {
			return array[this.front];
		}
		
	}

	@Override
	public E dequeue() {
		if (this.isEmpty()) {
			return null;
		} else {
			E elementToReturn = this.array[this.front];
			this.front = this.incrementIndex(this.front);
			this.size--;
			return elementToReturn;
		}
	}

	@Override
	public E dequeue(int index) throws NoSuchElementException {
		if (this.isEmpty()) {
			throw new NoSuchElementException("The queue is empty, no elements available");
		}
		
		if ((index < 0) || (index >= this.size)) {
			// not a valid index
			throw new NoSuchElementException("The specified index = " + index + " is not valid. The queue only has " + this.size + " elements.");
		}
		
		int arrayIndex = this.findIndex(index);
		E elementToReturn = this.array[arrayIndex];
		this.size--;
		
		System.out.println("Dequeue(" + index + ")");
		
		for ( int i = index, j = this.incrementIndex(i) ; j != this.back ; i = this.incrementIndex(i), j = this.incrementIndex(j) ) {
			this.array[this.findIndex(i)] = this.array[this.findIndex(j)];
		}
		this.back = this.decrementIndex(this.back);

		return elementToReturn;
	}

	@Override
	public void removeAll() {
		this.front = -1;
		this.back = -1;
		this.size = 0;
	}

	@Override
	public String toString() {
		String s = "ArrayBasedQueue = {";
		
		if ( this.size == 0) {
			s = s + "}";
		} else {
			for ( int i = this.front ; i != this.back ; i = this.incrementIndex(i)) {
				s = s + this.array[i] + ", ";
			}
			s = s + this.array[this.back] + "}";
		}
		s = s + " Size = " + this.size();
		return s;
	}

	/*************************************************************/
	/*     Private Methods										 */
	/*************************************************************/
	
	private int incrementIndex(int index) {
		return (index + 1) % this.array.length;
	}

	private int decrementIndex(int index) {
		int computedIndex = -1;
		
		computedIndex = index - 1;
		
		if (computedIndex < 0) {
			computedIndex = this.array.length - 1;
		}
		
		return computedIndex;
	}
	
	
	private int findIndex(int index) {
		int computedIndex = -1;
		
		if (!this.isEmpty() && index < this.size) {
			computedIndex = (this.front + index) % this.array.length;
		}
		
		return computedIndex;
	}
}
